import polars as pl
import subprocess
import orjson
import uuid
import rich
import copy
import sys
import os
import re

from pathlib import Path
from typing import List, Tuple, Dict, Union, Optional, Any, Iterator, TypeVar, MutableSequence

from iotree.core.render.funcs import rich_func_chainer
from iotree.core.render.tables import tableFromRecords

from ..utils.render import console, errors
from ..utils.paths import data_dir, config, assets_dir

iSqlColumns = Union[str, List[str]]
iSqlRows = Union[str, List[List[str]]]
iSqlRecord = Union[Dict[str, str]]
iSqlRecords = List[iSqlRecord]
iSqlResponse = Union[str, Tuple[iSqlColumns, iSqlRows]]
iSqlError = Dict[str, str]
iSqlPath = Union[str, os.PathLike, Path]
iSqlPathList = List[Union[str, os.PathLike, Path]]

class iSqlDF(pl.DataFrame):
    def __rich_repr__(self) -> rich.table.Table:
        records = orjson.loads(self.write_json(row_oriented=True))
        return tableFromRecords(records, theme="default")



def check_sql(txt):
    return (
        'select' in txt.lower() or
        'insert' in txt.lower() or
        'update' in txt.lower() or
        'delete' in txt.lower() or 
        'create' in txt.lower() or
        'drop' in txt.lower() or
        'alter' in txt.lower()
    )

def checkError(raw : str) -> Union[bool, None]:
    """Check if the raw output of isql contains an error message."""
    pattern = r"Msg (\d+), Level (\d+), State (\d+):\nServer '([\w\d\-_]+)', Line (\d+):"
    
    if groups := re.search(pattern, raw):
        base = {
            "msg": groups.group(1),
            "level": groups.group(2),
            "state": groups.group(3),
            "server": groups.group(4),
            "line": groups.group(5),
        }
        idx = groups.span()[1]
        reason = raw[idx:].strip()

        errors.print(f"[bold red]Couldn't parse error message.\n{orjson.dumps(base, option=orjson.OPT_INDENT_2).decode()}[/]\n[bold yellow]Reason: [/][yellow]{reason}[/]")
        sys.exit(1)

    else:
        if (
            "invalid" in raw.lower()
            and "level" in raw.lower()
            and "state" in raw.lower()
            and "line" in raw.lower()
            ):

            base = {
                "msg": None,
                "level": None,
                "state": None,
                "server": None,
                "line": None,
            }
            reason = "Couldn't parse reason, but error is present."
            errors.print(f"[bold red]Couldn't parse error message.\n{orjson.dumps(base, option=orjson.OPT_INDENT_2).decode()}[/]\n[bold yellow]Reason: [/][yellow]{reason}[/]")
            sys.exit(1)
        else:
            return False
    

    
def format_sql_file(filename: Union[str, os.PathLike, Path]) -> None:
    """Format a SQL file. isql doesn't accept scripts with no `GO` statements at the end of each query."""
    content = open(filename, "r").read().lower().strip()
    if not check_sql(content):
        return filename
    elif content.endswith("go"):
        return filename
    else:
        open(filename, "a").write("\nGO")
        return
    
class iSQL:
    def __init__(
        self, 
        server: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        db: Optional[str] = None
        ) -> None:

        default_config = config['default']

        self.server = server or default_config['server']
        self.user = user or default_config['user']
        self.password = password or default_config['password']
        self.db = db or default_config['db']

        self.preprocess = None
        
        if not self.server or not self.user or not self.password:
            raise ValueError("Must provide server, user, and password")
        
    def run_sql_file(
        self,
        filename: Union[str, os.PathLike, Path],
        preprocess: Optional[bool] = True,
        ) -> pl.DataFrame:
        """Run a query and return the result as a list of tuples."""
        
        self.preprocess = preprocess
        
        self.command = [
            'isql',
            "-S", self.server,
            "-U", self.user,
            "-P", self.password,
            "-s", "';'",
            "-D", self.db,
            "-J", "utf8",
            "-X", # encrypt pwd
            "-i", filename
        ]
        
        format_sql_file(filename) # format the file if necessary
        
        self.result = subprocess.run(
            self.command,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=subprocess.CREATE_NO_WINDOW,
            text=True,
            shell=True,
        )

        try:
            self.raw = self.result.stdout
            self.raw = self.decode('utf-8', errors='ignore') if isinstance(self.raw, bytes) else self.raw
        except Exception as err:
            raise err

        try:
            if preprocess:
                df = iSQL.strRowColSplitter(self.raw)
                return df
            else:
                return self.raw
        
        except Exception as err:
            raise err
        
    def query(
        self,
        query: str,
        preprocess: Optional[bool] = True,
        clean : Optional[bool] = True,
        ) -> Union[iSqlResponse, iSqlRecords, iSqlError]:
        """
        Run a query by storing it into a temporary .sql file
        and return the result as a list of tuples.
        """
        tmpdir = data_path.joinpath('tmp')
        tmpdir.mkdir(exist_ok=True)
        filename = tmpdir.joinpath(f"tmp{uuid.uuid4()}.sql")

        open(filename, "w+", encoding='utf-8').write(query)
        
        try:
            return self.run_sql_file(
                file = filename,
                preprocess = preprocess
            )
        except Exception as err:
            if clean:
                filename.unlink(missing_ok=True)
            raise err
        finally:
            filename.unlink(missing_ok=True)

    @staticmethod
    def strRowColSplitter(raw: str) -> List[str]:
        """Split a raw string output into rows and columns."""

        checkError(raw)
        endings = ["row affected)", "rows affected)"]
        for end in endings:
            idx = raw.rfind(end)
            if idx != -1:
                jdx = raw.rfind("(")
                raw = raw[:jdx]

        output = re.sub(r"( [ ]+)", " ", raw).strip()

        unicode = r"-"
        dbunicode = r"--"
        startcols = output.find(unicode)
        endcols = output.rfind(dbunicode)
        rawcols = output[:startcols]
        rawrows = output[endcols:]

        rawcols = re.sub(r"(\s+)", ";;", rawcols)
        cols = re.split(r"(;;|')", rawcols)
        cols = [ col for col in cols if col.strip() not in ["", ";;", "'"]]

        ko_rows = []
        ok_rows = []
        rawrows_aslist = rawrows.split("\'\n\'")
        rawrows_aslist = [ row for row in rawrows_aslist if set(row) != set("-") and row.strip() not in ["", "''"]]
        rawrows_aslist = list(map(lambda row: re.sub(r"(\n\t)", ";", row), rawrows_aslist))
        rawrows_aslist = list(map(lambda row: re.sub(r"(\d+)'", r"\1;", re.sub(r"(NULL')", "NULL;", row)), rawrows_aslist))

        if len(rawrows_aslist) == 1:
            row = rawrows_aslist[0]
            cols = cols if len(cols) else None
            if ";" in row and cols:
                return iSqlDF([{col: rs.strip() for col,rs in zip(cols,row.split(";")) if rs.strip() != ""}], orient="row")
            elif ";" in row and not cols:
                return iSqlDF([[rs.strip() for rs in row.split(";") if rs.strip() != ""]], orient="row")
            else:
                return iSqlDF([[row]], columns=cols[0], orient="row")

        result_rows = [ [v.strip() for v in row.split(";")] for row in rawrows_aslist ]
        df = iSqlDF(result_rows, columns=cols, orient="row")

        return df

    def multiquery(self, queries_source: Union[str, os.PathLike, Path]) -> List[iSqlDF]:
        """Run multiple queries and return the result as a list of pl.DataFrames"""
        if isinstance(queries_source, (Path, os.PathLike, str, list)):
            if not isinstance(queries_source, list):
                if Path(queries_source).is_dir():
                    queries = list(Path(queries_source).glob("*.sql"))
                else:
                    queries = [queries_source]
            if isinstance(queries_source, list):
                queries = []
                for query in queries_source:
                    if Path(query).is_dir():
                        queries += list(Path(query).glob("*.sql"))
                    else:
                        queries.append(query)
        else:
            raise TypeError(f"queries_source must be a string, a Path or a directory name, not {type(queries_source)}")
        callbacks = [self.run_sql_file for _ in range(len(queries))]
        return rich_func_chainer(callbacks, params=queries)

    def __repr__(self) -> str:
        return f"<iSQL: {self.server}>"
    
    def __str__(self) -> str:
        return self.__repr__()

    def __enter__(self):
        return self

    def copy(self, deep: Optional[bool] = False):
        return copy.deepcopy(self) if deep else copy.copy(self)

    def __or__(self, other):
        return self.copy().query(other)

    def __ror__(self, other):
        return self.copy().query(other)

    def __gt__(self, other):
        return self.copy().query(other)
    
    def __lt__(self, other):
        return self.copy().query(other)

    def __rshift__(self, queries_source: Union[str, os.PathLike, Path]) -> List[pl.DataFrame]:
        """Pipes a list of .sql files or a directory name to the pisql 🐍 command 🦾"""
        return self.multiquery(queries_source)

    def __rshift__(self, queries_source: Union[str, os.PathLike, Path]) -> List[pl.DataFrame]:
        """Pipes a list of .sql files or a directory name to the pisql 🐍 command 🦾"""
        return self.multiquery(queries_source)
        


import random
import orjson
import typer
import rich
import sys
import os
import re

from pathlib import Path

from typing import List, Tuple, Dict, Union, Optional, Any, Iterator, TypeVar

from iotree.core.render.tables import tableFromRecords
from iotree.core.render.funcs import basic_pbar

from pisql.core.isql import iSQL
from pisql.utils.render import console, errors

from pisql.utils.paths import (
    base_dir, tests_dir, package_dir,
    config, assets_dir, data_dir
)



isql = iSQL(**config["default"])

app = typer.Typer(
    name="pisql",
    help="A simple but rich command line interface for Sybase ASE",
    no_args_is_help=True,
    rich_help_panel="rich",
    rich_markup_mode="rich",
)

exe = typer.Typer(
    name="execute",
    help=f"""Execute a query ❓ against a Sybase ASE database 🧑‍💻. Accepts a .sql file 📄",
    [yellow][bold]Aliases:[/] exec, e, x[/]  
    [green][bold]Example:[/]  
    >>> pisql x name.sql[/]
    """,
    rich_markup_mode="rich",
    rich_help_panel="rich",
)

dev = typer.Typer(
    name="dev",
    help=f"""Developer tools 💻.
    [yellow][bold]Aliases:[/] d, dev[/]
    """,
    rich_markup_mode="rich",
    rich_help_panel="rich",
    no_args_is_help=True,
)

@exe.callback(invoke_without_command=True, no_args_is_help=True)
def executer(
    file: str = typer.Argument(..., help="The .sql file 📄 to execute"),
    out: str = typer.Option(None, "-o", "--out", "--out-file", help="The output file 📰, defaults to rich table 📊 on console"),
    override_default_dir: str = typer.Option(None, "-od", "--override-default-dir", help="Override the default directory 📁. Defaults to `~/pisql` 😉"),
    ) -> None:
    """Execute a query ❓ against a Sybase ASE database 📦. Accepts a .sql file 📄"""

    file = Path(file)

    if not override_default_dir:
        dir = Path(os.path.expanduser("~")) / "pisql" / "results"
        dir.mkdir(parents=True, exist_ok=True)
    else:
        dir = Path(override_default_dir) if override_default_dir != "." else Path.cwd()
        dir.mkdir(parents=True, exist_ok=True)

    with basic_pbar() as pbar:
        task = pbar.add_task(f"Executing {file.name}", total=None)
        df = isql.run_sql_file(file)

        pbar.update(task, advance=50)
        pbar.update(task, description=f"Writing results to {dir}")

        csv_file = dir / "{file.stem}.csv"
        parquet_file = dir / "{file.stem}.parquet"

        if out is None:
            csv_out = df.write_csv(file=csv_file)
            parquet_out = df.write_parquet(file=parquet_file,  compression="lz4")
            records = orjson.loads(df.write_json(row_oriented=True))
            console.print(tableFromRecords(records, theme="default"))
        else:
            outfile = dir / file.name
            csv_out = df.write_csv(file=csv_file)
            parquet_out = df.write_parquet(file=parquet_file, compression="lz4")
            if out.lower() in ["json", "excel"]:
                errors.print(f"[bold red]Output format {out} not supported yet with the base version.[/]", style="bold red")
                errors.print(f"[dim yellow]Please install `pandas` to get excel support[/]")
                sys.exit(1)
        pbar.update(task, advance=50)
        pbar.update(task, description=f"✅ Donew writing results to {dir}", completed=True)
        sys.exit(0)

@exe.command(name=">>>", help="Execute a query ❓ against a Sybase ASE database 📦. Accepts a dir 📁 of .sql files 📄 or a list of files")
def runall(
    args : List[Path],
    dir: str = typer.Option(None, "-d", "--dir", help="the base directory of all the files")
    ) -> None:
    """Execute a query ❓ against a Sybase ASE database 📦. Accepts a dir 📁 of .sql files 📄 or a list of files"""
    files = []

    if dir is not None:
        dir = Path(dir)
        if dir.is_dir():
            files += [file for file in dir.iterdir() if file.suffix == ".sql"]
        else:
            errors.print(f"[bold red]Invalid directory {dir}[/]", style="bold red")
            sys.exit(1)
    
    for arg in args:
        arg = Path(arg)
        if arg.is_dir():
            files += [file for file in arg.iterdir() if file.suffix == ".sql"]
        elif arg.is_file():
            files.append(arg)
        else:
            errors.print(f"[bold red]Invalid file/dir {arg}[/]", style="bold red")
            sys.exit(1)
    for file in files:
        executer(file=file)


def listTablesServer() -> None:
    """List all tables 📓 in the server 📦"""
    querytxt = """
        SELECT t.name, c.name
        FROM sysobjects t, syscolumns c
        WHERE t.id = c.id AND t.type IN ('U', 'V')
        """
    df = isql.query(querytxt)
    console.print(df)
    df.write_parquet(pisql_home / "results" / "tables.parquet", compression="lz4")

@exe.command(name="tables?", help="List all tables 📓 in the server 📦. [bold yellow]Alias: `?`[/]")
def listTables() -> None:
    """List all tables 📓 in the server 📦"""
    listTablesServer()

@exe.command(name="?", help="List all tables 📓 in the server 📦")
def listTables() -> None:
    """List all tables 📓 in the server 📦"""
    listTablesServer()


@dev.command(name="test", help="Run tests 🧪")
def test(
    test_type: str = typer.Argument(..., help="The type of test to run 🧪"),
    index: int = typer.Option(None, "-i", "-idx", "--index", help="The index of the subtest to run [dim](if your test has many options)[/] 🧪"),
    name: str = typer.Option(None, "-n", "--name", help="The name (a part of it) of the subtest to run [dim](if your test has many options)[/] 🧪"),
    ) -> None:
    """Run tests 🧪"""
    test_type = test_type.lower()
    if index is not None and name is not None:
        errors.print("[bold red]You can't specify both index and name[/]", style="bold red")
        errors.print("[dim yellow]Using `name`.[/]")
        index = None
        selector = name
    elif index is not None:
        selector = index
    elif name is not None:
        selector = name
    else:
        selector = None
    
    if test_type in ["parse", "parsing", "text"]:
        raw_outs = {
            name: open( data_dir / name, "r").read() for name in os.listdir(data_dir)
        }
        if selector is None:
            selector = random.randint(0, len(raw_outs) - 1)

        target = [
            name for name in raw_outs.keys() if selector in name
        ]

        if not len(target):
            raise typer.BadParameter(f"Couldn't find a test with name {selector}")

@dev.command(name="view", help="See the raw data 📰 🧪")
def test(
    name: str = typer.Argument(..., help="The name (a part of it) of the subtest to run [dim](if your test has many options)[/] 🧪"),
    head: int = typer.Option(500, "-h", "--head", help="The number of lines to show [dim](if your file is huge !)[/] 🧪"),
    raw: bool = typer.Option(False, "-ra", "--raw", help="Whether to show the raw data 📰 🧪"),
    remove_whitespaces: bool = typer.Option(False, "-rw", "--rm-ws", "--remove-whitespaces", help="Remove whitespaces from the output [dim](if your data uses them too much !)[/] 🧪"),
    columns: bool = typer.Option(False, "-c", "--cols", "--columns", help="Whether to only display the parsed columns 🏦"),
    ) -> None:

    raw_outs = {
            name: open( data_dir / name, "r").read() for name in os.listdir(data_dir)
        }
    target = [
        key for key in raw_outs.keys() if name in key
    ]

    if not len(target):
        raise typer.BadParameter(f"Couldn't find a raw datafile with name {name}")
    
    target = target.pop()
    raw_out = raw_outs[target]

    if raw:
        idx = 0
        for i in range(head):
            idx = raw_out.find("'", idx + 1)
        console.print(raw_out[:idx])
        sys.exit(0)

    if columns:
        df = isql.strRowColSplitter(raw_out)
        console.print(df.columns)
        sys.exit(0)
    else:
        df = isql.strRowColSplitter(raw_out)
        records = orjson.loads(df.write_json(row_oriented=True))
        console.print(tableFromRecords(records[:head], theme="default"))

    if remove_whitespaces:
        raw_out = re.sub(r"([ \t\n]+)", " ", raw_out)
    
    sys.exit(0)



@dev.command(name="list", help="List the test queries 📰🗄️")
def list():
    """List the test queries 📰🗄️"""
    queries_dir = assets_dir / "sql"
    console.print(f"[bold magenta]Test queries:[/]")
    
    for file in os.listdir(queries_dir):
        console.print(f"[yellow] \u2736 📰 {file}[/]")
    
    console.print(f"[bold magenta]Total:[/] {len(os.listdir(queries_dir))} 📰🗄️")
    console.print(f"Let's go to work ! 🚀🚀🚀")

app.add_typer(exe, name="execute")
app.add_typer(exe, name="exec", hidden=True)
app.add_typer(exe, name="e", hidden=True)
app.add_typer(exe, name="x", hidden=True)

app.add_typer(dev, name="dev", no_args_is_help=True)
app.add_typer(dev, name="d", no_args_is_help=True, hidden=True)




import polars as pl
import typer

from typing import List, Tuple, Dict, Union, Optional, Any, Iterator, TypeVar
from pisql.core.isql import iSQL
from pisql.utils.paths import (
    base_dir, tests_dir, package_dir
)

qapp = typer.Typer(
    name="query",
    help="Run a query against a Sybase ASE database",
    no_args_is_help=True,
)

@qapp.command()
def query(
    query: typer.Argument(..., help="The query to run"),
    ):
    pass


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pisql', 'pisql.cli', 'pisql.core', 'pisql.utils']

package_data = \
{'': ['*'], 'pisql': ['assets/*', 'assets/data/*', 'assets/sql/*']}

install_requires = \
['iotree>=0.1.12,<0.2.0', 'polars>=0.15.13,<0.16.0', 'typer[all]>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['pisql = pisql.main:app']}

setup_kwargs = {
    'name': 'pisql',
    'version': '0.1.10',
    'description': '',
    'long_description': '',
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
